# -*- python -*-
# $RCSfile: adaptivemeshrefinement.py,v $
# $Revision: 1.17 $
# $Author: langer $
# $Date: 2006/03/24 22:27:03 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 


from oof2.common import registeredclass
from oof2.common.IO import parameter
from oof2.common.IO import xmlmenudump
from oof2.engine import errorestimator
from oof2.engine import meshmod
from oof2.engine import rationalize
from oof2.engine import refine
from oof2.engine import skeletonelement
from oof2.engine import skeletonboundary
from oof2.engine import skeletonmodifier
from oof2.SWIG.common import switchboard
from oof2.SWIG.common.IO import progressbar


SkeletonEdgeBoundary = skeletonboundary.SkeletonEdgeBoundary
ExteriorSkeletonEdgeBoundary = skeletonboundary.ExteriorSkeletonEdgeBoundary
SkeletonPointBoundary = skeletonboundary.SkeletonPointBoundary
ExteriorSkeletonPointBoundary = skeletonboundary.ExteriorSkeletonPointBoundary

# AdaptiveMeshRefinement
class AdaptiveMeshRefinement(meshmod.MeshModification):
    def __init__(self, estimator, criterion, degree, alpha,
                 rationalize=0):
        self.estimator = estimator  # ErrorEstimator instance
        self.criterion = criterion  # RefinementCriterion instance
        self.degree = degree  # RefinementDegree instance
        self.alpha = alpha
        self.rationalize = rationalize
        
    def doTheBoundary(self, old_skel, new_skel):
        # Point boundary propagation.
        for pbdy in old_skel.sheriffSkeleton().pointboundaries.values():
            pbdy.makeContextBoundary(
                None, pbdy.get_name(), old_skel).map(old_skel, new_skel)
        # Edge boundary propagation.
        for ebdy in old_skel.sheriffSkeleton().edgeboundaries.values():
            ebdy.makeContextBoundary(
                None, ebdy.get_name(), old_skel).map(old_skel, new_skel)
        
    def apply(self, meshcontext):
        pBar = progressbar.getProgress()
        
        # refine the skeleton.
        refinstance = refine.Refine(
            errorestimator.AdaptiveMeshRefine(meshcontext.path(),
                                              self.estimator),
            self.criterion, self.degree, self.alpha)
        old_skel = meshcontext.getSkeleton()
        new_skel = refinstance.applyAMR(old_skel, None)
            
        # Boundary propagation.
        self.doTheBoundary(old_skel, new_skel)
        # Only the second Skeleton need to be disconnected.
        if meshcontext.skeleton_buffer.isFirstInLine(old_skel):
            new_skel.disconnect()

        if pBar.query_stop():
            pBar.set_failure()
            pBar.set_message("Failed")
            
            return None
        else:
            pBar.set_progress(0.5)
            pBar.set_message("Mesh Refined / Post-processing.")
        
        # rationalize the skeleton
        if self.rationalize:
            newer_skel = new_skel.properCopy(fresh=True)
            targets = skeletonmodifier.AllElements()
            criterion = skeletonmodifier.AverageEnergy(0.8)  # alpha
            for ratreg in rationalize.Rationalizer.registry:
                # Create a Rationalizer from a Registration
                # with default values.
                ratmethod = ratreg()
                ratmethod(newer_skel, None, targets, criterion,
                          ratmethod.findAndFix)
            newer_skel.cleanUp()
            self.doTheBoundary(new_skel, newer_skel)
            new_skel = newer_skel

        meshcontext.pushSkeleton(new_skel)

        # convert the refined skeleton into an femesh
        old_femesh = meshcontext.femesh()
        # realmaterial is a function which tells the elements of the
        # created femesh how to get their materials.
        realmaterial = skeletonelement.SkeletonElement.realmaterial
        new_femesh = new_skel.femesh(meshcontext.elementdict,
                                     realmaterial)
        new_femesh.set_rwlock(meshcontext.rwLock)

        # copy field status
        for field in meshcontext.all_fields():
            new_femesh.define_field(field)
            new_femesh.activate_field(field)
            if old_femesh.in_plane(field):
                field.set_in_plane(new_femesh, 1)
        
        # copy equation status
        for equation in meshcontext.all_equations():
            new_femesh.activate_equation(equation)

        # replace old_femesh
        meshcontext.setFemesh(new_femesh)

        # putting BC in new_Femesh
        for (name, bc) in meshcontext.allBoundaryConds():
            bc.boundary_obj=new_femesh.getBoundary(bc.boundary)
            new_femesh.boundaries[bc.boundary].addCondition(bc)

        # Fields to be initialized from old to new.
        for field in meshcontext.all_fields():
            new_femesh.init_field(old_skel, old_femesh, field)
        ## transfer time index
        new_femesh.set_time(old_femesh.current_time())

        # destroying old_femesh
        old_femesh.destroy()
                
        # time stamp update
        meshcontext.solution_is_current = 0
        meshcontext.changed()

        if pBar.query_stop():
            pBar.set_failure()
            pBar.set_message("Failed")
            return None
        else:
            pBar.set_progress(1.)
            pBar.set_message("Done.")

registeredclass.Registration(
    'Adaptive Mesh Refinement',
    meshmod.MeshModification,
    AdaptiveMeshRefinement,
    ordering=0,
    params=[
    parameter.RegisteredParameter('estimator',
                                  errorestimator.ErrorEstimator,
                                  tip="Which error estimator to use."),
    parameter.RegisteredParameter('criterion',
                                  refine.RefinementCriterion,
                                  tip="Only consider elements meeting the criterion."),
    parameter.RegisteredParameter('degree',
                                  refine.RefinementDegree,
                                  tip="How to subdivide elements."),
    skeletonmodifier.alphaParameter,
    parameter.BooleanParameter('rationalize', 0, default=0,
                               tip='Rationalize mesh?')
    ],
    tip="Adaptively refine mesh based on error estimation.",
    discussion=xmlmenudump.loadFile('DISCUSSIONS/engine/reg/adaptive.xml')
    )
