// -*- C++ -*-
// $RCSfile: cskeleton.h,v $
// $Revision: 1.40 $
// $Author: rlua $
// $Date: 2006/07/17 15:58:15 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#ifndef CSKELETON_H
#define CSKELETON_H

#include "common/boolarray.h"
#include "common/coord.h"
#include "common/timestamp.h"
#include "common/pixelsetboundary.h"
#include <vector>
#include <iostream>

class CSkeletonNode;
class CSkeletonElement;
class CSkeleton;

class MasterCoord;
class CMicrostructure;
class TwoMatrix;

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

class CSkeletonNode {
private:
  unsigned char mobility;
  static const unsigned char xmovable_;
  static const unsigned char ymovable_;
  static const unsigned char unpinned_;
  Coord position_;
  Coord lastposition_;
  TimeStamp lastmoved;
public:
  CSkeletonNode(double x, double y)
    : mobility(xmovable_|ymovable_|unpinned_),
      position_(x, y)
  {}
  const Coord &position() const { return position_; }
  TimeStamp nodemoved;
//   bool movable_x() const { return mobility & xmovable_; }
//   bool movable_y() const { return mobility & ymovable_; }
  bool movable_x() const {
    return (mobility & xmovable_) && (mobility & unpinned_);
  }
  bool movable_y() const {
    return (mobility & ymovable_) && (mobility & unpinned_);
  }
  bool pinned() const { return !(mobility & unpinned_); }
  bool movable() const { return movable_x() || movable_y(); }
  void setMobilityX(bool mob);
  void setMobilityY(bool mob);
  void setPinned(bool pin);
  void copyMobility(const CSkeletonNode *other) {
    mobility = other->mobility;
  }
  bool canMergeWith(const CSkeletonNode*) const;
  bool moveTo(const Coord *pos); // takes pointer so that SWIG typemap can work
  void unconstrainedMoveTo(const Coord *pos);
  bool moveBy(const Coord *shift);
  void moveBack();
//   Coord random_position(double, double) const;
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

typedef std::multimap<double, PixelBdyIntersection> ElEdgeMap;
typedef std::vector<ElEdgeMap> ElEdge;
typedef std::pair<double, PixelBdyIntersection> ElEdgeDatum;

typedef std::multimap<Coord, PixelBdyIntersection> CoordIsec;
typedef std::pair<Coord, PixelBdyIntersection> CoordIsecDatum;

// Utility function for deleting a PixelBdyIntersection from
// the node-index-indexed lists.
void delete_isec(const PixelBdyIntersection&, ElEdge &);

class HomogeneityData {
private:
  double homogeneity;
  int dominantpixel;
  double homog_energy;
  // This is the "clone" constructor.
  HomogeneityData(double hom, int cat, double nrg) : homogeneity(hom),
						     dominantpixel(cat),
						     homog_energy(nrg) {}
public:
  HomogeneityData(double hom, int cat);
  HomogeneityData() : homogeneity(0), dominantpixel(0), homog_energy(0) {}
  double get_homogeneity() const { return homogeneity; }
  int get_dominantpixel() const { return dominantpixel; }
  double get_energy() const { return homog_energy; }
  HomogeneityData clone() { return HomogeneityData(this->homogeneity,
						   this->dominantpixel,
						   this->homog_energy); }
};

class CSkeletonElement {
protected:
  std::vector<CSkeletonNode*> nodes;
  virtual double shapefun(int, const MasterCoord&) const = 0;
public:
  CSkeletonElement(int n) 
    : nodes(n) 
  {}
  virtual ~CSkeletonElement() {}
  int nnodes() const { return nodes.size(); }
  void replaceNode(int which, CSkeletonNode *replacement);
  std::vector<Coord> *perimeter() const; // list of node positions
  double perimeterLength() const;
  double perimeterSquared() const;
  double edgeLength(int) const;
  double edgeLengthSquared(int) const;
  double cosCornerAngle(int) const;
  double cosCornerAngleSquared(int) const;
  double getRealAngle(int) const;
  Coord frommaster(MasterCoord *mc, int rotation) const; // convert master coord
  Coord center() const;		// average position of nodes
  CRectangle bbox() const;

  // The first "interior" function is valid for all elements, and only
  // takes a single argument, but can give ambiguous results for edge
  // cases.  The second one takes an additional perturbation-hint, and
  // will only work (but is quite robust) for convex elements.
  virtual bool interior(const Coord*) const = 0; // is the given point inside?
private:
  bool interior(const Coord&, const Coord&) const; 

public:
  virtual double area() const = 0;
  virtual bool illegal() const = 0;
  virtual const std::vector<ICoord> *underlying_pixels(const CMicrostructure&)
    const = 0;
  double pixel_intersection(const ICoord *pxl, const Coord *pxlsize) const;
  const std::vector<double> *categoryAreas(CMicrostructure&) const;
  HomogeneityData c_homogeneity(CMicrostructure&) const;
  bool transitionPoint(CMicrostructure&, int, Coord*) const;
  double edgeHomogeneity(CMicrostructure&, int) const;
  virtual double energyShape() const = 0;
};

class CSkeletonTriangle : public CSkeletonElement {
protected:
  virtual double shapefun(int, const MasterCoord&) const;
public:
  CSkeletonTriangle(CSkeletonNode*, CSkeletonNode*, CSkeletonNode*);
  virtual bool illegal() const;
  virtual bool interior(const Coord*) const;
  virtual double area() const;
  virtual const std::vector<ICoord> *underlying_pixels(const CMicrostructure&)
    const;
  virtual double energyShape() const;
};

class CSkeletonQuad : public CSkeletonElement {
protected:
  virtual double shapefun(int, const MasterCoord&) const;
public:
  CSkeletonQuad(CSkeletonNode*, CSkeletonNode*, CSkeletonNode*, CSkeletonNode*);
  virtual bool illegal() const;
  virtual bool interior(const Coord*) const;
  virtual double area() const;
  virtual const std::vector<ICoord> *underlying_pixels(const CMicrostructure&)
    const;
  virtual double energyShape() const;
};

//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// class CSkeleton {
// private:
//   std::vector<CSkeletonNode*> nodes;
//   std::vector<CSkeletonElement*> elements;
// public:
//   CSkeleton();
// };

#endif // CSKELETON_H

